#ifndef INFO_BASE
#define INFO_BASE

#include <list.h>
#include <time.h>
#include <MAPLAB/IO/string_to_width.h>
#include <MAPLAB/basic.h>

template <class T>
struct CoordXDA
{
  typedef T value_type;
};

template <class T, class It>
inline T get(CoordXDA<T> da, It &it)
{
  return (*it).x;
}
template <class T>
struct CoordYDA
{
  typedef T value_type;
};

template <class T, class It>
inline T get(CoordYDA<T> da, It &it)
{
  return (*it).y;
}
template <class T>
struct NameDA
{
  typedef T value_type;
};

template <class T, class It>
inline T get(NameDA<T> da, It &it)
{
  return (*it).my_site->name;
}
template <class T>
struct SiteNameDA
{
  typedef T value_type;
};

template <class T, class It>
inline T get(SiteNameDA<T> da, It &it)
{
  return (*it).name;
}
template <class T>
struct CandBeginDA
{
  typedef T value_type;
};

template <class T, class It>
inline T get(CandBeginDA<T> da, It &it)
{
  return (*it).candidate_list.begin();
}
template <class T>
struct CandEndDA
{
  typedef T value_type;
};

template <class T, class It>
inline T get(CandEndDA<T> da, It &it)
{
  return (*it).candidate_list.end();
}
template <class T>
struct LLXDA
{
  typedef T value_type;
};

template <class T, class It>
inline T get(LLXDA<T> da, It &it)
{
  return (*it).llx;
}
template <class T>
struct LLYDA
{
  typedef T value_type;
};

template <class T, class It>
inline T get(LLYDA<T> da, It &it)
{
  return (*it).lly;
}
template <class T>
struct URXDA
{
  typedef T value_type;
};

template <class T, class It>
inline T get(URXDA<T> da, It &it)
{
  return (*it).urx;
}
template <class T>
struct URYDA
{
  typedef T value_type;
};

template <class T, class It>
inline T get(URYDA<T> da, It &it)
{
  return (*it).ury;
}

template <class T>
struct WidthDA
{
  typedef T value_type;
};

template <class T, class It>
inline void Set(WidthDA<T> da, It const it, T width)
{
  (*it).width = width;
}

template <class CoordType, class NameType>
class Information_base
{
public:
  typedef CoordType value_type;
  typedef NameType name_type;
private:
  struct Site;
  friend struct Candidate;
  struct Candidate
  {
    Site* my_site;
    value_type llx, lly, urx, ury;
    Candidate(Site* site, value_type ll_x, value_type ll_y,
              value_type ur_x, value_type ur_y)
    : my_site(site), llx(ll_x), lly(ll_y), urx(ur_x), ury(ur_y)
    {}
  
    void trans(value_type dx, value_type dy)
    {
      llx += dx; lly += dy; urx += dx; ury += dy;
    }
  
    void scale(double fx, double fy)
    {
      llx *= fx; lly *= fy; urx *= fx; ury *= fy;
    }
  };
  friend struct Site;
  struct Site
  {
    value_type x, y;
    list<Candidate> candidate_list;
    bool name_set;
    name_type name;
    value_type width, height;
    value_type four_cand_width, four_cand_hght;
  
    Site(value_type x_coord, value_type y_coord)
    : x(x_coord), y(y_coord), name_set(false)
    {}
  
    void insert_candidate(value_type low_x, value_type low_y,
                          value_type high_x, value_type high_y)
    {
      candidate_list.insert(candidate_list.end(),
                            Candidate(this, low_x, low_y, high_x, high_y));
    }
  
    void trans(value_type dx, value_type dy)
    {
      x += dx; y += dy;
      Candidate_it begin, end =  candidate_list.end();
      for (begin = candidate_list.begin(); begin != end; begin++)
        (*begin).trans(dx, dy);
    }
  
    void scale(double fx, double fy)
    {
      x *= fx; y *= fy;
      Candidate_it begin, end =  candidate_list.end();
      for (begin = candidate_list.begin() ; begin != end; begin++)
        (*begin).scale(fx, fy);
    }
  
    void insert_name_candidates()
    {
      insert_candidate(x - width, y - height, x, y);
      insert_candidate(x, y - height, x + width, y);
      insert_candidate(x, y, x + width, y + height);
      insert_candidate(x - width, y, x, y + height);
    }
  };
  typedef list<Site> Sites;
  typedef list<Candidate> Candidates;
  /*
  friend class Window_filter;
  class Window_filter
  {
  public:
    Window_filter(value_type ll_x, value_type ll_y,
                  value_type ur_x, value_type ur_y)
    : llx(ll_x), lly(ll_y), urx(ur_x), ury(ur_y)
    {}
  
    Window_filter()
    {}
  
    bool operator() (Site& const site)
    {
      if ((llx > site.x) || (lly > site.y) ||
          (urx < site.x) || (ury < site.y))
        return false;
      return true;
    }
  private:
    value_type llx, lly, urx, ury;
  };
  */
public:
  typedef typename Sites::iterator Site_it;
  typedef typename Candidates::iterator Candidate_it;
  // typedef filter_it<Site_it, Site, Window_filter> Site_window_it;
  typedef CoordXDA<value_type> CoordXDA;
  typedef CoordYDA<value_type> CoordYDA;
  typedef NameDA<name_type> NameDA;
  typedef SiteNameDA<name_type> SiteNameDA;
  typedef CandBeginDA<Candidate_it> CandBeginDA;
  typedef CandEndDA<Candidate_it> CandEndDA;
  typedef LLXDA<value_type> LLXDA;
  typedef LLYDA<value_type> LLYDA;
  typedef URXDA<value_type> URXDA;
  typedef URYDA<value_type> URYDA;
  
  Site_it sites_begin()
  {
    return sites.begin();
  }
  
  Site_it sites_end()
  {
    return sites.end();
  }
  
  Candidate_it cand_begin()
  {
    return all_candidates.begin();
  }
  
  Candidate_it cand_end()
  {
    return all_candidates.end();
  }
  
  int number_of_sites()
  {
    return sites.size();
  }
  
  int number_of_cands()
  {
    return all_candidates.size();
  }
  /*
  Site_window_it sites_window_begin(value_type ll_x, value_type ll_y,
                                    value_type ur_x, value_type ur_y)
  {
    return Site_window_it(sites.begin(), sites.end(),
                          Window_filter(ll_x, ll_y, ur_x, ur_y) );
  }
  
  Site_window_it sites_window_end()
  {
    return Site_window_it(sites.end());
  }
  
  int number_of_sites_in_window(value_type ll_x, value_type ll_y,
                                value_type ur_x, value_type ur_y)
  {
    int size = 0;
    Site_window_it current(sites.begin(), sites.end(),
                           Window_filter(ll_x, ll_y, ur_x, ur_y) );
    Site_window_it end(sites.end());
  
    while (current != end) size++;
    return size;
  }
  */
  
  void store_to_file(String &file_name,
                     String &creator,
                     String &version,
                     String &labelpositions = "",
                     CoordType width = 0,
                     CoordType height = 0)
  {
    ofstream out_file(file_name);
    if (file_name.contains('/')) file_name = file_name.after('/', -1);
    time_t tm = time(0);
    out_file << "%Filename: " << file_name << endl
             << "%Creator: " << creator << "-" << version << endl
             << "%Date: " << ctime(&tm);
  
    Site_it end = sites.end();
    if (atoi(labelpositions.chars()) == 4)
    {
      out_file << "%Labelpositions: 4" << endl;
      if (!((width == 0) && (height == 0)))
      {
        out_file << "%Labelsize: " << width << " " << height << endl;
        for ( Site_it current = sites.begin(); current != end; current++)
          out_file << (*current).x << " " << (*current).y << endl;
      }
      else
      {
        for ( Site_it current = sites.begin(); current != end; current++)
        {
            out_file << (*current).x << " " << (*current).y << " ";
            if ((*current).name_set)
              out_file << (*current).width << " " << (*current).height
                       << endl;
            else
              out_file << (*current).four_cand_width << " "
                       << (*current).four_cand_hght << endl;
        }
      }
    }
    else
    {
      for ( Site_it current = sites.begin(); current != end; current++)
      {
        out_file << (*current).x << " " << (*current).y << " ";
        Candidate_it cand_end = (*current).candidate_list.end(),
          cand_current;
        for (cand_current = (*current).candidate_list.begin();
             cand_current != cand_end; cand_current++)
          out_file << "(" << (*cand_current).llx << " "
                   << (*cand_current).lly << " " << (*cand_current).urx
                   << " " << (*cand_current).ury << ")";
        out_file << endl;
      }
    }
  }
protected:
  Site_it insert_site(value_type x, value_type y)
  {
    if (sites.size() == 0)
    {
      xmin = x; ymin = y; xmax = x; ymax = y;
      max_width = 0; max_height = 0;
    }
    else
    {
      xmin = minimum(xmin, x); ymin = minimum(ymin, y);
      xmax = maximum(xmax, x); ymax = maximum(ymax, y);
    }
    return sites.insert(sites.end(), Site(x, y));
  }
  void insert_label(Site_it site, value_type low_x, value_type low_y,
                    value_type high_x, value_type high_y)
  {
    xmin = minimum(xmin, low_x); ymin = minimum(ymin, low_y);
    xmax = maximum(xmax, high_x); ymax = maximum(ymax, high_y);
    (*site).insert_candidate(low_x, low_y, high_x, high_y);
    all_candidates.push_back(Candidate(&(*site), low_x, low_y, high_x, high_y));
  }
  void insert_four_labels(Site_it site, value_type width, value_type height)
  {
    (*site).four_cand_width = width;
    (*site).four_cand_hght = height;
    value_type x = (*site).x;
    value_type y = (*site).y;
    insert_label(site, x - width, y - height, x, y);
    insert_label(site, x, y - height, x + width, y);
    insert_label(site, x, y, x + width, y + height);
    insert_label(site, x - width, y, x, y + height);
  }
  void set_name(Site_it site, name_type& n)
  {
    (*site).name_set = true;
    (*site).name = n;
  }
  void set_width_height(String& fontname, int fontsize)
  {
    Site_it end = sites.end();
    string_to_width(sites.begin(), end, String("gs"), fontname,
                    fontsize, sites.begin(), SiteNameDA(),
                    WidthDA<value_type>());
  
    max_height = fontsize;
    for ( Site_it current = sites.begin(); current != end; current++)
    {
      if (max_width < (*current).width) max_width = (*current).width;
      (*current).height = fontsize;
    }
  }
  void insert_name_candidates()
  {
    Site_it end = sites.end();
  
    for ( Site_it begin = sites.begin(); begin != end; begin++)
      if ((*begin).name_set) (*begin).insert_name_candidates();
  }
  
  void make_bounding_box(value_type b_llx, value_type b_lly,
                         value_type b_urx, value_type b_ury)
  {
    value_type llx = b_llx + max_width;
    value_type lly = b_lly + max_height;
    value_type urx = b_urx - max_width;
    value_type ury = b_ury - max_height;
  
    translate(llx, lly);
  
    double dx = xmax - xmin;
    double dy = ymax - ymin;
    dx = ( dx ? dx : 1 );
    dy = ( dy ? dy : 1 );
    double fx = (urx - llx) / dx;
    double fy = (ury - lly) / dy;
    double f  = minimum(fx, fy);
  
    scale(f, f);
    translate(llx, lly);
  }
  void make_window(value_type w_llx, value_type w_lly,
                         value_type w_urx, value_type w_ury)
  {
    Site_it end = sites.end();
    Site_it current = sites.begin(), curr_save;
  
    while (current != end)
    {
      curr_save = current;
      current++;
      if ((w_llx > (*curr_save).x) || (w_urx < (*curr_save).x) ||
          (w_lly > (*curr_save).y) || (w_ury < (*curr_save).y))
        sites.erase(curr_save);
    }
  
    for (current = sites.begin(); current != end; current++)
    {
      if (current == sites.begin())
      {
        xmin = xmax = (*current).x;
        ymin = ymax = (*current).y;
        max_width = (*current).width;
        max_height = (*current).height;
      }
      else
      {
        xmin = minimum(xmin, (*current).x); ymin = minimum(ymin,
                                                            (*current).y);
        xmax = maximum(xmax, (*current).x); ymax = maximum(ymax,
                                                            (*current).y);
        max_width = maximum(max_width, (*current).width);
        max_height = maximum(max_height, (*current).height);
      }
    }
  }
  void translate(value_type b_llx, value_type b_lly)
  {
    double dx = b_llx - xmin;
    double dy = b_lly - ymin;
  
    Site_it end = sites.end();
    for ( Site_it begin = sites.begin(); begin != end; begin++)
      (*begin).trans(dx, dy);
  
    xmin += dx; ymin += dy; xmax += dx; ymax += dy;
  }
  
  void scale(double fx, double fy)
  {
    Site_it end = sites.end();
    for (Site_it begin = sites.begin(); begin != end; begin++)
      (*begin).scale(fx, fy);
  
    xmin *= fx; ymin*= fy; xmax *= fx; ymax *= fy;
  }
private:
  Sites sites;
  Candidates all_candidates;
  value_type xmin, ymin, xmax, ymax, max_width, max_height;
};

#endif
