#include <map.h>
#include <set.h>
#include <strstream.h>
#include <String.h>
class Make_statistics
{
  enum {MAX = 1024} ;
  typedef pair<int, double> Qual_time_value;
  typedef map<String, Qual_time_value, less<String> > Algo_value_map;
  typedef typename Algo_value_map::value_type Algo_value;
  typedef typename Algo_value_map::iterator Algo_it;
  typedef map<String, Algo_value_map, less<String> > Table;
  typedef typename Table::value_type Table_value;
  typedef typename Table::iterator Table_it;
  typedef set<String, less<String> > Algo_set;
  typedef typename Algo_set::iterator Algo_set_it;
public:

  Make_statistics(String& Creator, String& optimal_algo = String("Total"))
  : my_optimal_algo(optimal_algo), creator(Creator)
  {}

  void insert_in_table(String& filename)
  {
    char buff[MAX];
    ifstream datafile(filename);
    String algo = filename.after('/', -1);
    algo = algo.before('_');
    algo_set.insert(algo);

    while (datafile.get(buff, MAX))
    {
      char new_line_char;
      datafile.get(new_line_char);
      String line(buff);
      if (!line.empty() && isalpha(line.firstchar()))
      {
        String example_file = line.before(' ');
        line = line.after(' ');
        istrstream ist(line.chars(), line.length());
        int num_of_pt, num_of_labeled_pt;
        double time_taken;
        ist >> num_of_pt >> num_of_labeled_pt >> time_taken;
        if (num_of_labeled_pt == 0) num_of_labeled_pt = 10;
        Table_it table_it = my_table.find(example_file);
        if (table_it != my_table.end())
          (*table_it).second.insert(Algo_value(algo,
                                     Qual_time_value(num_of_labeled_pt,
                                                     time_taken)));
        else
        {
          Algo_value_map algo_value_map;
          algo_value_map.insert(
            Algo_value(algo, Qual_time_value(num_of_labeled_pt,
                                             time_taken)));
          algo_value_map.insert(Algo_value(String("Total"),
                                      Qual_time_value(num_of_pt, 0)));
          my_table.insert(Table_value(example_file, algo_value_map));
        }
      }
    }
  }

  void save_quality_time(String& destination)
  {
    Algo_set_it end = algo_set.end();
    for (Algo_set_it begin = algo_set.begin(); begin != end; begin++)
      save_quality_time(String(*begin), destination);
  }

  void save_quality_time(String& algo, String& destination)
  {
    Table_it end = my_table.end();
    Table_it begin = my_table.begin();
    while (begin != end)
    {
      String example_file = (*begin).first;
      String example_class = example_file.before("_");
      ofstream quality_result(destination + "/" + example_class + "_" +
                              algo + ".qual");
      ofstream time_result(destination + "/" + example_class + "_" + algo +
                           ".time");
      make_headers(quality_result, algo, example_class);
      quality_result << "#Format: average_number  average_quality" <<
        "  min_quality max_quality" << endl << endl;
      make_headers(time_result, algo, example_class);
      time_result << "#Format: average_number  average_time  min_time"
      "  max_time" << endl << endl;

      while ((begin != end) && example_file.contains(example_class))
      {
        String example_class_size = example_file.before("_", -1);
        int num_of_examples = 0, total_points = 0;
        double total_quality = 0, max_quality = 0, min_quality = 10000,
          total_time = 0, max_time = 0, min_time, time;
        while ((begin != end) && example_file.contains(example_class_size))
        {
          double quality = get_quality(begin, algo, total_points, time);
          if (quality != 0)
          {
            num_of_examples++;

            total_quality += quality;
            if (quality < min_quality) min_quality = quality;
            if (quality > max_quality) max_quality = quality;

            total_time += time;
            if ((num_of_examples == 1) || (time < min_time))
              min_time = time;
            if (time > max_time) max_time = time;
          }
          ++begin;
          if (begin != end)
          {
            String file_buffer = (*begin).first;
            example_class_size = file_buffer.before("_", -1);
          }
        }
        if (num_of_examples != 0)
        {
          double average_pts =
          (double)total_points/(double)num_of_examples;
          double average_quality =

          // this line has been changed due to rectangle intersection
          total_quality/(double)num_of_examples; // *100;
          double average_time = total_time/(double)num_of_examples;

          quality_result.precision(2);
          quality_result.setf(ios::fixed,ios::floatfield);
          quality_result.setf(ios::right,ios::adjustfield);
          quality_result << average_pts << "  " << average_quality << "  "
             << (min_quality) << "  " << (max_quality) << endl;

          time_result.precision(2);
          time_result.setf(ios::fixed,ios::floatfield);
          time_result.setf(ios::right,ios::adjustfield);
          time_result << average_pts << "  " << average_time << "  " <<
            min_time << "  " << max_time << endl;
        }
        if (begin != end)
          example_file = (*begin).first;
      }
    }
  }

  void set_optimal_algo(String& opt_algo)
  {
    my_optimal_algo = opt_algo;
  }

  double get_quality(Table_it& table_it, String& algo, int& total_points,
                     double& time)
  {
    double quality = 0;
    Algo_it algo_it = (*table_it).second.find(algo);
    if (algo_it != (*table_it).second.end())
    {
      int pts_labeled = (*algo_it).second.first;
      time = (*algo_it).second.second;

      Algo_it total_algo_it = (*table_it).second.find("Total");
      int opt_pts_labeled = (*total_algo_it).second.first;;
      if (my_optimal_algo == algo)
      {
        total_points += opt_pts_labeled;
        int buff = opt_pts_labeled;
        opt_pts_labeled = pts_labeled;
        pts_labeled = buff;
      }
      else if (my_optimal_algo != String("Total"))
      {
        Algo_it opt_algo_it = (*table_it).second.find(my_optimal_algo);
        if (opt_algo_it != (*table_it).second.end())
        {
          total_points += opt_pts_labeled;
          opt_pts_labeled = (*opt_algo_it).second.first;
        }
        else pts_labeled = 0;
      }
      else total_points += opt_pts_labeled;
      // this line has been changed due to rectangle intersection
      quality= ((double)pts_labeled);// /((double)opt_pts_labeled);
    }
    return quality;
  }

  void make_headers(ofstream& file, String& algo, String& example_class)
  {
    time_t tm = time(0);
    file << "#Creator: " << creator << endl <<
      "#Solver: " << algo << endl <<
      "#Generator: " << example_class << endl <<
      "#Date: " << ctime(&tm) <<
      "#Platform: " << getenv("MAKE_PLATFORM") << endl <<
      "#Host: " << getenv("HOST") << endl << endl;
}

private:
  Table my_table;
  String my_optimal_algo, creator;
  Algo_set algo_set;
};
