#line 315 "conflicts_compare.aw"

// using LEDA timer
#include <LEDA/basic.h>

// algorithms
#include <gen_rec_inters.h>
#include <oo_rec_inters.h>
// standard libraries
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <assert.h>
#include <String.h>

#include <MAPLAB/error_handling/assertion.h>

// auxiliary functions
#include <MAPLAB/basic.h>  // for convert_to_str()
#include <MAPLAB/IO/file_reader.h>

#line 82 "conflicts_compare.aw"
typedef File_reader<double, String> My_file_reader;
typedef typename My_file_reader::value_type Coord_type;
typedef typename My_file_reader::Candidate_it Cand_it;
typedef typename My_file_reader::LLXDA Ll_x_da;
typedef typename My_file_reader::LLYDA Ll_y_da;
typedef typename My_file_reader::URXDA Ur_x_da;
typedef typename My_file_reader::URYDA Ur_y_da;
#line 335 "conflicts_compare.aw"
                                               
#line 95 "conflicts_compare.aw"
// OO interface
typedef OORecInters<double> OO_rec_algo;
typedef typename OO_rec_algo::Rectangle_ls Rectangle_ls;
typedef typename OO_rec_algo::Rectangle_ls_it Rectangle_ls_it;
typedef typename OO_rec_algo::Graph Conflict_graph;
typedef typename Conflict_graph::Node_it Node_it;
typedef typename Conflict_graph::Adj_edge_it Adj_edge_it;

class Rec_it_comp
{
public:
  bool operator() (Rectangle_ls_it& r1, Rectangle_ls_it& r2)
  { return true; }
};
typedef multimap<Rectangle_ls_it, list<Rectangle_ls_it>, Rec_it_comp >
  My_graph_1;
typedef typename My_graph_1::value_type My_graph_pair_1;

// Generic interface
typedef GenRecInters< Cand_it, Ll_x_da, Ll_y_da, Ur_x_da, Ur_y_da >
Gen_rec_algo;
typedef typename Gen_rec_algo::Rectangle_it Gen_rec_algo_it;
class Gen_rec_it_comp
{
public:
  bool operator() (Gen_rec_algo_it& r1, Gen_rec_algo_it& r2)
  { return true; }
};
typedef multimap<Gen_rec_algo_it, list<Gen_rec_algo_it>, Gen_rec_it_comp >
  My_graph_2;
typedef typename My_graph_2::value_type My_graph_pair_2;

#line 336 "conflicts_compare.aw"


#line 130 "conflicts_compare.aw"
main(int argc, char *argv[])
{
  if (argc < 3)
  {
    cout << "usage:  make_protocol {-m} {-t{<sec>}} " << endl
         << "        <name of algorithm> <file1> { ... <fileN> }"
         << " <destination directory>" << endl;
    exit(-1);
  }

  int index = 0, time_limit = -1;
  bool ps_output = false, send_mail = false;
  String ps_dir(".");
  if (String(argv[index+1]) == String("-m")) index++;
  index++;
  String algo(argv[index]);
  String destination(argv[argc-1]);

  time_t tm = time(0);
  String result(ctime(&tm));
  result = result.after(' ');
  result.gsub(' ', '_');
  result = algo + "_" + result.before('\n');
  result = destination +
    (destination[destination.length()-1] == '/' ? "" : "/") + result;
  String result2 = result + "(Conflicts.)";
  ofstream datafile1(result);
  ofstream datafile2(result2);
  assert(datafile1);
  assert(datafile2);
  const int MAXLEN = 255;
  char buffer[MAXLEN];

  datafile1 << "#Solver: " << algo << endl
           << "#Date: " << ctime(&tm)
           << "#Platform: " << getenv("MAKE_PLATFORM") << endl
           << "#Host: " << getenv("HOST") << "(" << getenv("HOSTTYPE")
           << ")" << endl
           << "#Format:  file name            #rectangles  #conflicts"
           << "     CPU secs" << endl;
 datafile2 << "#Solver: " << algo << endl
           << "#Date: " << ctime(&tm)
           << "#Platform: " << getenv("MAKE_PLATFORM") << endl
           << "#Host: " << getenv("HOST") << "(" << getenv("HOSTTYPE")
           << ")" << endl
           << "#Format:  file name            #rectangles  #conflicts"
           << "     CPU secs" << endl;
  cout     << " #Format:  file name             #rectangles   #conflicts"
           << "     CPU secs" << endl;

  for (int i = index + 1; i < (argc-1); i++)
  {
    ostrstream data(buffer, MAXLEN);
    ostrstream data2(buffer, MAXLEN);
    assert(data);
    assert(data2);

    String file_name = argv[i];
    cout << " processing " << file_name; cout.flush();
    My_file_reader my_reader((const char*) file_name);
    cout << "..." << endl; cout.flush();
    int num_of_cand = my_reader.number_of_cands();
    int conflicts = 0;
    float sec, sec2 = 0;

    if (algo == "OORecInters")
    {
      float T = used_time();
      Rectangle_ls rectangles1, rectangles2;
      Cand_it curr = my_reader.cand_begin();
      Cand_it end = my_reader.cand_end();
      for ( ; curr != end; curr++)
        rectangles1.push_back(Rectangle<double>(get(Ll_x_da(), curr),
                                       get(Ll_y_da(), curr),
                                       get(Ur_x_da(), curr),
                                       get(Ur_y_da(), curr)));
      OO_rec_algo my_algo_1(rectangles1);
      my_algo_1.run();
      conflicts = my_algo_1.num_of_conflicts();
      sec2 = used_time(T);

      T = used_time();
      curr = my_reader.cand_begin();
      end = my_reader.cand_end();
      for ( ; curr != end; curr++)
        rectangles2.push_back(Rectangle<double>(get(Ll_x_da(), curr),
                                       get(Ll_y_da(), curr),
                                       get(Ur_x_da(), curr),
                                       get(Ur_y_da(), curr)));
      OO_rec_algo my_algo_2(rectangles2);
      Conflict_graph* graph = my_algo_2.run();
      My_graph_1 my_graph_1;
      Node_it node_current = graph->nodes_begin();
      Node_it node_end = graph->nodes_end();
      for (int i = 0; node_current != node_end; ++node_current, i++)
      {
        list<Rectangle_ls_it> result;
        Adj_edge_it edge_current =
        graph->adj_edges_begin(node_current);
        Adj_edge_it edge_end = graph->adj_edges_end(node_current);
        for ( ; edge_current != edge_end; ++edge_current)
          result.push_back(*(graph->opposite(node_current,
                         graph->get_edge_it(edge_current))));
        my_graph_1.insert(My_graph_pair_1(*node_current, result));
      }
      sec = used_time(T);
    }
    else if (algo == "GenRecInters")
    {
      float T = used_time();
      Gen_rec_algo my_algo_1(my_reader.cand_begin(),
                             my_reader.cand_end());
      conflicts = my_algo_1.num_of_conflicts();
      sec2 = used_time(T);

      T = used_time();
      Gen_rec_algo my_algo_2(my_reader.cand_begin(),
                             my_reader.cand_end());
      My_graph_2 my_graph_2;
      while (my_algo_2.valid())
      {
        if (my_algo_2.step() == Gen_rec_algo::rectangle_start)
        {
          list<Gen_rec_algo_it> result;
          Gen_rec_algo_it current = my_algo_2.begin();
          Gen_rec_algo_it end = my_algo_2.begin();
          for ( ; !(current == end); ++current)
            result.push_back(current);
          my_graph_2.insert(My_graph_pair_2(current, result));
        }
      }
      sec = used_time(T);
    }
    file_name = file_name.after(file_name.index("/", -1));
    data.width(25);
    data.setf(ios::left,ios::adjustfield);
    data << file_name;
    data.width(11);
    data.setf(ios::right,ios::adjustfield);
    data << num_of_cand;
    data.width(11);
    data << conflicts;
    data.width(13);
    data.precision(2);
    data.setf(ios::fixed,ios::floatfield);
    data.setf(ios::right,ios::adjustfield);
    data << sec << endl << '\0';

    datafile1 << buffer;
    datafile1.flush();
    cout << " " << buffer;

    data2.width(25);
    data2.setf(ios::left,ios::adjustfield);
    data2 << file_name;
    data2.width(11);
    data2.setf(ios::right,ios::adjustfield);
    data2 << num_of_cand;
    data2.width(11);
    data2 << conflicts;
    data2.width(13);
    data2.precision(2);
    data2.setf(ios::fixed,ios::floatfield);
    data2.setf(ios::right,ios::adjustfield);
    data2 << sec2 << endl << '\0';

    datafile2 << buffer;
    datafile2.flush();
  }

  if (send_mail)
  {
    String message = String("echo 'process ") + String(argv[0]) +
      String(" on '$HOST' terminated' | mail $LOGNAME");
    system(message.chars());
  }
  tm = time(0);
  datafile1 << "#Terminated: " << ctime(&tm) << endl;
  datafile2 << "#Terminated: " << ctime(&tm) << endl;
}
#line 338 "conflicts_compare.aw"
 
